/******************************************************************************
*
* project name:    TI-Chess
* file name:       interrupt.c
* initial date:    12/07/2000
* authors:         thomas.nussbaumer@gmx.net (coding)
*                  marcos.lopez@wol.es (design/graphics/beta testing)
* description:     keyboard handler routine installed as int1 handler
*
* !!! Thanx to Zeljko Juric to help me out with informations !!!
*
* $Id: interrupt.c,v 1.2 2000/08/12 15:31:13 Thomas Nussbaumer Exp $
*
******************************************************************************/


//-----------------------------------------------------------------------------
// One major problem is that the keyboard of the TI calculators is NOT
// unbounced! So don't believe in short transitions !!
//
// NOTE: Also the settle time of the register inquiry for _rowread() MUST be
//       increased by a factor of 3 for HW 2 calculators
//-----------------------------------------------------------------------------

#include "interrupt.h"

volatile unsigned char onkey     = 0;  // may used to detect [ON] key pressed for break
volatile unsigned int  do_reset  = 0;  // may be used to clear transition buffer
volatile unsigned char o_onkey   = 0;  // previous on key state
volatile unsigned char n_onkey   = 0;  // new on key state
volatile unsigned int  row2scan  = 0;  // actual row to scan
unsigned char          trow[10]  = {}; // may be used to detect key-down transitions
unsigned char          orow[10]  = {}; // previous key states
unsigned char          nrow[10]  = {}; // actual key states
unsigned long          old_int1_handler = 0; // backup variable for old int1 handler
unsigned long          old_int2_handler = 0; // backup variable for old int2 handler
int                    is_hw1_calc = 0;  // will be set to 1 if calculator is HW1
                                         // by InstallInterruptHandlers()

//-----------------------------------------------------------------------------
// useful macros to turn off and on interrupts
//-----------------------------------------------------------------------------
#define IntsOff() {asm("move.l %d0,-(%sp);move.w #0x0700,%d0;trap #1;move.l (%sp)+,%d0");}
#define IntsOn()  {asm("move.l %d0,-(%sp);move.w #0x0000,%d0;trap #1;move.l (%sp)+,%d0");}



//-----------------------------------------------------------------------------
// definitions of vector table entry addresses
//-----------------------------------------------------------------------------
#define VTABLE_ENTRY_INT1 0x64
#define VTABLE_ENTRY_INT2 0x68
#define VTABLE_ENTRY_INT3 0x6C
#define VTABLE_ENTRY_INT4 0x70
#define VTABLE_ENTRY_INT5 0x74
#define VTABLE_ENTRY_INT6 0x78
#define VTABLE_ENTRY_INT7 0x7C



/*===========================================================================*/
/* sets new interrupt handler                                                */
/*                                                                           */
/*===========================================================================*/
unsigned long GetHardwareVersion() {
  unsigned long hwpb, *rombase;
  rombase = (unsigned long *)((*(unsigned long *)0xC8) & 0x600000);
  hwpb = rombase[65];
  return (hwpb - (unsigned long)rombase < 0x10000 &&
    *(unsigned short *)hwpb > 22 ? *(unsigned long *)(hwpb + 22) : 1);
}



/*===========================================================================*/
/* sets new interrupt handler                                                */
/*                                                                           */
/* parameters:                                                               */
/*                                                                           */
/* handler_addr - address of new handler given as simple unsigned long       */
/* vec_addr     - address of the vector table entry which should be used     */
/*                                                                           */
/* return value:                                                             */
/*                                                                           */
/* address of previous handler routine as plane unsigned long                */
/*                                                                           */
/* NOTE: this routine is used for both: installing and deinstalling our      */
/*       new interrupt handler                                               */
/*                                                                           */
/*===========================================================================*/
unsigned long SetHandler(unsigned long handler_addr,unsigned long vec_addr) {
    unsigned long old_int;

    IntsOff();
    old_int = *(unsigned long*)vec_addr;
    vec_addr+=0x40000;
    *(unsigned long*)(vec_addr+0x40000) = handler_addr;
    IntsOn();
    return old_int;
}


//-----------------------------------------------------------------------------
// use higher settle time for register query to fix problems with 
// HW2 calculators (redefinition of TIGCC macro)
//-----------------------------------------------------------------------------
#undef _rowread
#define _rowread(x) ({register int __r;*(int*)0x600018=(int)(x); \
  asm volatile("move.w #24,%0;0:dbra %0,0b":"=g"(__r)); \
  ~(*(unsigned char*)0x60001b);})



/*===========================================================================*/
/* our keyboard new keyboard handler detects and stores key down transitions */
/* (used internally)                                                         */
/*                                                                           */
/* this handler uses a round robin-like method, which means that at each     */
/* call just one row is scanned to keep the time effort for each call small  */
/*===========================================================================*/
void KeyboardHandler(void) {
    if (do_reset) {
        memset(trow,0,10);
        do_reset = 0;
        onkey    = 0;
    }

    nrow[row2scan] = _rowread(~(1<<row2scan));
    trow[row2scan] |= (~orow[row2scan]) & nrow[row2scan];
    orow[row2scan] = nrow[row2scan];

    row2scan++;
    if (!TI89) {
       if (row2scan == 10) row2scan = 0;
    }
    else {
       if (row2scan == 7)  row2scan = 0;
    }

    //-------------------------------------
    // get state of [ON] key
    //-------------------------------------
    n_onkey = ~(*(unsigned char*)0x60001a) & 0x2;

    if (!o_onkey && n_onkey) onkey = 1;
    o_onkey = n_onkey;
}



/*===========================================================================*/
/* this is used as wrapper for our keyboard handler                          */
/* (just used internally)                                                    */
/*===========================================================================*/
void IKeyWrapper(void);  // just the prototype
asm("IKeyWrapper:
     movem.l %d0-%d7/%a0-%a6,-(%sp)
     bsr KeyboardHandler
     movem.l (%sp)+,%d0-%d7/%a0-%a6
     rte");



/*===========================================================================*/
/* this is a dummy key handler which gets installed for INT2                 */
/* (just used internally)                                                    */
/*===========================================================================*/
void DummyKeyHandler(void);  // just the prototype
asm("DummyKeyHandler: rte");



/*===========================================================================*/
/* installs new INT handlers                                                 */
/*                                                                           */
/* INT1 handler == own keyboard handler                                      */
/* INT2 handler == dummy handler to increase performance                     */
/*===========================================================================*/
void InstallInterruptHandlers() {
    unsigned long hwpb, *rombase,result;

    rombase = (unsigned long *)((*(unsigned long *)0xC8) & 0x600000);
    hwpb = rombase[65];
    result = (hwpb - (unsigned long)rombase < 0x10000 &&
    *(unsigned short *)hwpb > 22 ? *(unsigned long *)(hwpb + 22) : 1);

    if (result == 1) is_hw1_calc = 1;
    else             is_hw1_calc = 0;

    //---------------------------------------
    // initialize globals of keyboard handler
    //---------------------------------------
    onkey    = 0;
    o_onkey  = 0;
    n_onkey  = 0;
    do_reset = 0;
    row2scan = 0;
    memset(orow,0,10);
    memset(nrow,0,10);
    memset(trow,0,10);
    //---------------------------------------
    // install handler for INT1 and INT2
    //---------------------------------------
    old_int1_handler = SetHandler((unsigned long)IKeyWrapper,VTABLE_ENTRY_INT1);
    old_int2_handler = SetHandler((unsigned long)DummyKeyHandler,VTABLE_ENTRY_INT2);
}



/*===========================================================================*/
/* restores old INT handlers                                                 */
/*===========================================================================*/
void RestoreInterruptHandlers() {
    SetHandler(old_int1_handler,VTABLE_ENTRY_INT1);
    SetHandler(old_int2_handler,VTABLE_ENTRY_INT2);
}



//#############################################################################
//###################### NO MORE FAKES BEYOND THIS LINE #######################
//#############################################################################
//
//=============================================================================
// Revision History
//=============================================================================
//
// $Log: interrupt.c,v $
// Revision 1.2  2000/08/12 15:31:13  Thomas Nussbaumer
// substitution keywords added
//
//
